# Copyright 2020 The GNOME Music developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

from __future__ import annotations
from typing import Optional, Union
import asyncio
import typing

import gi
gi.require_versions({"MediaArt": "2.0", "Soup": "3.0"})
from gi.repository import Gio, GLib, GObject, MediaArt, Soup, GdkPixbuf

from gnomemusic.musiclogger import MusicLogger
from gnomemusic.utils import CoreObjectType
if typing.TYPE_CHECKING:
    from gnomemusic.corealbum import CoreAlbum
    from gnomemusic.coreartist import CoreArtist
    from gnomemusic.coresong import CoreSong

if typing.TYPE_CHECKING:
    CoreObject = Union[CoreAlbum, CoreArtist, CoreSong]


class StoreArt(GObject.Object):
    """Stores Art in the MediaArt cache.
    """

    __gsignals__ = {
        "finished": (GObject.SignalFlags.RUN_FIRST, None, ())
    }

    _sem = asyncio.Semaphore(10)

    def __init__(self):
        """Initialize StoreArtistArt

        :param coreobject: The CoreArtist or CoreAlbum to store art for
        :param string uri: The art uri
        """
        super().__init__()

        self._coreobject = None

        self._file = None
        self._log = MusicLogger()
        self._soup_session = Soup.Session.new()

    async def _start(
            self, coreobject: CoreObject, uri: str,
            coreobjecttype: CoreObjectType) -> None:
        self._coreobject = coreobject

        if uri in [None, ""]:
            self.emit("finished")
            return

        album: Optional[str] = None
        artist = self._coreobject.props.artist
        art_prefix = "artist"
        success = False

        if coreobjecttype == CoreObjectType.ALBUM:
            album = self._coreobject.props.title
            art_prefix = "album"
        elif coreobjecttype == CoreObjectType.SONG:
            album = self._coreobject.props.album
            art_prefix = "album"

        success, self._file = MediaArt.get_file(artist, album, art_prefix)

        if not success:
            self.emit("finished")
            return

        async with self._sem:
            cache_dir = GLib.build_filenamev(
                [GLib.get_user_cache_dir(), "media-art"])
            cache_dir_file = Gio.File.new_for_path(cache_dir)

            try:
                await cache_dir_file.query_info_async(
                    Gio.FILE_ATTRIBUTE_ACCESS_CAN_READ,
                    Gio.FileQueryInfoFlags.NONE, GLib.PRIORITY_DEFAULT_IDLE)
            except GLib.Error:
                # directory does not exist yet
                try:
                    cache_dir_file.make_directory(None)
                except GLib.Error as error:
                    self._log.warning(
                        "Error: {}, {}".format(error.domain, error.message))
                    self.emit("finished")
                    return

            msg = Soup.Message.new("GET", uri)
            try:
                gbytes = await self._soup_session.send_and_read_async(
                    msg, GLib.PRIORITY_DEFAULT)
            except GLib.Error as error:
                self._log.debug(
                    f"Failed to get remote art: {error.domain}, "
                    f"{error.message} for album: {album} by artist: {artist}")
                self.emit("finished")
                return

            istream = Gio.MemoryInputStream.new_from_bytes(gbytes)
            try:
                pixbuf = await GdkPixbuf.Pixbuf.new_from_stream_async(istream)
            except GLib.Error as error:
                self._log.warning(f"Error: {error.domain}, {error.message}")
                self.emit("finished")
                return
            finally:
                await istream.close_async(GLib.PRIORITY_DEFAULT)

            try:
                ostream = await self._file.create_async(
                    Gio.FileCreateFlags.NONE, GLib.PRIORITY_DEFAULT_IDLE)
            except GLib.Error as error:
                # File already exists.
                self._log.info(f"Error: {error.domain}, {error.message}")
                self.emit("finished")
                return
            else:
                try:
                    _, buffer = pixbuf.save_to_bufferv("jpeg")
                except GLib.Error as error:
                    self._log.warning(
                        f"Error: {error.domain}, {error.message} for album: "
                        f"{album} by artist: {artist}")
                    await ostream.close_async(GLib.PRIORITY_DEFAULT_IDLE)
                    await self._file.delete_async(GLib.PRIORITY_DEFAULT_IDLE)
                    self.emit("finished")
                    return

                try:
                    await ostream.write_async(
                        buffer, GLib.PRIORITY_DEFAULT_IDLE)
                except GLib.Error as error:
                    self._log.info(f"Error: {error.domain}, {error.message}")
                    self.emit("finished")
                    return

                self._coreobject.props.thumbnail = self._file.get_uri()
                await ostream.close_async(GLib.PRIORITY_DEFAULT)

        self.emit("finished")

    def start(
            self, coreobject: CoreObject, uri: str,
            coreobjecttype: CoreObjectType) -> None:
        """Start storing the art from the given URI

        :param coreobject: Any of CoreSong, CoreAlbum or CoreArtist
            the URI is relevant for
        :param str uri: The URI containing the art
        :param CoreObjectType coreobjecttype: The type of the
            coreobject
        """
        asyncio.create_task(self._start(coreobject, uri, coreobjecttype))



# Copyright 2024 The GNOME Music developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from __future__ import annotations
import typing

from gi.repository import GObject, Gio, Gtk, Adw

if typing.TYPE_CHECKING:
    from gnomemusic.application import Application


@Gtk.Template(resource_path='/org/gnome/Music/ui/PreferencesDialog.ui')
class PreferencesDialog(Adw.PreferencesDialog):
    """Main menu preferences dialog
    """

    __gtype_name__ = "PreferencesDialog"

    _inhibit_suspend_row = Gtk.Template.Child()
    _repeatmode_row = Gtk.Template.Child()
    _replay_gain_row = Gtk.Template.Child()

    def __init__(self, application: Application) -> None:
        """Initialize the application settings dialog.

        :param Application application: Application object
        """
        super().__init__()

        self._settings = application.props.settings

        self._repeatmode_row.props.selected = self._settings.get_enum("repeat")
        self._repeatmode_row.connect(
            "notify::selected", self._update_repeate_mode)

        self._replay_gain_row.props.selected = self._settings.get_enum(
            "replaygain")
        self._replay_gain_row.connect(
            "notify::selected", self._update_replaygain)

        self._settings.bind(
            "inhibit-suspend", self._inhibit_suspend_row, "active",
            Gio.SettingsBindFlags.DEFAULT)

    def _update_repeate_mode(
            self, row: Adw.ComboRow, value: GObject.ParamSpecInt) -> None:
        self._settings.set_enum("repeat", row.props.selected)

    def _update_replaygain(
            self, row: Adw.ComboRow, value: GObject.ParamSpecInt) -> None:
        self._settings.set_enum("replaygain", row.props.selected)



# Copyright 2024 The GNOME Music developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from __future__ import annotations
from enum import IntEnum
import typing

from gettext import gettext as _
from gi.repository import Adw, GLib, GObject, Gtk, Tracker

from gnomemusic.widgets.headerbar import HeaderBar
if typing.TYPE_CHECKING:
    from gnomemusic.application import Application


@Gtk.Template(resource_path="/org/gnome/Music/ui/StatusNavigationPage.ui")
class StatusNavigationPage(Adw.NavigationPage):
    """Status page when there is no music or an issue with Tracker

    This view can have several states

    EMPTY: No music has been found at startup (default)
    NO_TRACKER: Tracker is unavailable
    TRACKER_OUTDATED: Tracker version is too old
    """

    class State(IntEnum):
        """Enum for StatusNavigationPage state."""
        EMPTY = 0
        SEARCH = 1
        NO_TRACKER = 2
        TRACKER_OUTDATED = 3

    __gtype_name__ = "StatusNavigationPage"

    _description_label = Gtk.Template.Child()
    _initial_state = Gtk.Template.Child()
    _status_page = Gtk.Template.Child()
    _toolbar = Gtk.Template.Child()

    def __init__(self, application: Application) -> None:
        super().__init__()

        # FIXME: This is now duplicated here and in GrlTrackerWrapper.
        try:
            music_folder = GLib.get_user_special_dir(
                GLib.UserDirectory.DIRECTORY_MUSIC)
            assert music_folder is not None
        except (TypeError, AssertionError):
            self._content_text = _("Your XDG Music directory is not set.")
        else:
            music_folder = Tracker.sparql_escape_string(
                GLib.filename_to_uri(music_folder))
            href_text = "<a href='{}'>{}</a>".format(
                music_folder, _("Music Folder"))
            # TRANSLATORS: This is a label to display a link to open
            # a user's music folder. {} will be replaced with the
            # translated text 'Music folder'
            folder_text = _("The contents of your {} will appear here.")
            self._content_text = folder_text.format(href_text)

        self._headerbar = HeaderBar(application)
        self._headerbar.props.state = HeaderBar.State.EMPTY
        self._toolbar.add_top_bar(self._headerbar)

        # Hack to get to AdwClamp, so it can be hidden for the
        # initial state.
        child_of_child = self._status_page.get_first_child().get_first_child()
        self._adw_clamp = child_of_child.get_first_child().get_first_child()

        self._status_page.set_child(self._initial_state)

        self.props.state = StatusNavigationPage.State.EMPTY

    @GObject.Property(type=int, default=0, minimum=0, maximum=2)
    def state(self) -> int:
        """Get the state of the empty view

        :returns: The view state
        :rtype: int
        """
        return self._state

    @state.setter  # type: ignore
    def state(self, value: int) -> None:
        """Set the state of the empty view

        :param int value: new state
        """
        self._state = value

        self._adw_clamp.props.visible = True
        self._initial_state.props.visible = False

        if self._state == StatusNavigationPage.State.EMPTY:
            self._set_empty_state()
        elif self._state == StatusNavigationPage.State.SEARCH:
            self._set_search_state()
        elif self._state == StatusNavigationPage.State.NO_TRACKER:
            self._set_no_tracker_state()
        elif self._state == StatusNavigationPage.State.TRACKER_OUTDATED:
            self._set_tracker_outdated_state()

        self._headerbar.props.state = HeaderBar.State.EMPTY

    def _set_empty_state(self) -> None:
        self._adw_clamp.props.visible = False
        self._initial_state.props.visible = True

        self._description_label.props.label = self._content_text

    def _set_search_state(self) -> None:
        self._status_page.props.title = _("No Music Found")
        self._status_page.props.description = _("Try a Different Search")

    def _set_no_tracker_state(self) -> None:
        self._status_page.props.title = _(
            "GNOME Music could not connect to Tracker.")
        self._status_page.props.description = _(
            "Your music files cannot be indexed without Tracker running.")

        self._status_page.props.icon_name = "dialog-error-symbolic"

    def _set_tracker_outdated_state(self) -> None:
        self._status_page.props.title = _(
            "Your system Tracker version seems outdated.")
        self._status_page.props.description = _(
            "Music needs Tracker version 3.0.0 or higher.")

        self._status_page.props.icon_name = "dialog-error-symbolic"



# Copyright 2019 The GNOME Music developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

from __future__ import annotations
import asyncio
import re
import typing

from gi.repository import Gio, GLib

from gnomemusic.grilowrappers.grltrackerplaylists import Playlist
from gnomemusic.gstplayer import Playback
from gnomemusic.player import RepeatMode
from gnomemusic.widgets.songwidget import SongWidget

if typing.TYPE_CHECKING:
    from gi.repository import GObject

    from gnomemusic.coresong import CoreSong
    from gnomemusic.player import Player


class DBusInterface:

    def __init__(self, name, path, application):
        """Etablish a D-Bus session connection

        :param str name: interface name
        :param str path: object path
        :param GtkApplication application: The Application object
        """
        self._con: Gio.DBusConnection
        self._log = application.props.log
        self._path = path
        self._signals = None

        asyncio.create_task(self._get_bus(name))

    async def _get_bus(self, name: str) -> None:
        try:
            self._con = await Gio.bus_get(Gio.BusType.SESSION, None)
        except GLib.Error as error:
            self._log.warning(
                f"Unable to connect to the session bus: {error.message}")
            return

        Gio.bus_own_name_on_connection(
            self._con, name, Gio.BusNameOwnerFlags.NONE, None, None)

        method_outargs = {}
        method_inargs = {}
        signals = {}
        for interface in Gio.DBusNodeInfo.new_for_xml(self.__doc__).interfaces:

            for method in interface.methods:
                method_outargs[method.name] = "(" + "".join(
                    [arg.signature for arg in method.out_args]) + ")"
                method_inargs[method.name] = tuple(
                    arg.signature for arg in method.in_args)

            for signal in interface.signals:
                args = {arg.name: arg.signature for arg in signal.args}
                signals[signal.name] = {
                    'interface': interface.name, 'args': args}

            self._con.register_object(
                self._path, interface, self._on_method_call, None, None)

        self._method_inargs = method_inargs
        self._method_outargs = method_outargs
        self._signals = signals

    def _on_method_call(
        self, connection, sender, object_path, interface_name, method_name,
            parameters, invocation):
        """GObject.Closure to handle incoming method calls.

        :param Gio.DBusConnection connection: D-Bus connection
        :param str sender: bus name that invoked the method
        :param srt object_path: object path the method was invoked on
        :param str interface_name: name of the D-Bus interface
        :param str method_name: name of the method that was invoked
        :param GLib.Variant parameters: parameters of the method invocation
        :param Gio.DBusMethodInvocation invocation: invocation
        """
        args = list(parameters.unpack())
        for i, sig in enumerate(self._method_inargs[method_name]):
            if sig == 'h':
                msg = invocation.get_message()
                fd_list = msg.get_unix_fd_list()
                args[i] = fd_list.get(args[i])

        method_snake_name = DBusInterface.camelcase_to_snake_case(method_name)
        try:
            result = getattr(self, method_snake_name)(*args)
        except ValueError as e:
            invocation.return_dbus_error(interface_name, str(e))
            return

        # out_args is at least (signature1). We therefore always wrap the
        # result as a tuple.
        # Reference:
        # https://bugzilla.gnome.org/show_bug.cgi?id=765603
        result = (result,)

        out_args = self._method_outargs[method_name]
        if out_args != '()':
            variant = GLib.Variant(out_args, result)
            invocation.return_value(variant)
        else:
            invocation.return_value(None)

    def _dbus_emit_signal(self, signal_name, values):
        if self._signals is None:
            return

        signal = self._signals[signal_name]
        parameters = []
        for arg_name, arg_signature in signal['args'].items():
            value = values[arg_name]
            parameters.append(GLib.Variant(arg_signature, value))

        variant = GLib.Variant.new_tuple(*parameters)
        self._con.emit_signal(
            None, self._path, signal['interface'], signal_name, variant)

    @staticmethod
    def camelcase_to_snake_case(name):
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
        return '_' + re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


class MPRIS(DBusInterface):
    """
    <!DOCTYPE node PUBLIC
    '-//freedesktop//DTD D-BUS Object Introspection 1.0//EN'
    'http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd'>
    <node>
        <interface name='org.freedesktop.DBus.Introspectable'>
            <method name='Introspect'>
                <arg name='data' direction='out' type='s'/>
            </method>
        </interface>
        <interface name='org.freedesktop.DBus.Properties'>
            <method name='Get'>
                <arg name='interface' direction='in' type='s'/>
                <arg name='property' direction='in' type='s'/>
                <arg name='value' direction='out' type='v'/>
            </method>
            <method name='Set'>
                <arg name='interface_name' direction='in' type='s'/>
                <arg name='property_name' direction='in' type='s'/>
                <arg name='value' direction='in' type='v'/>
            </method>
            <method name='GetAll'>
                <arg name='interface' direction='in' type='s'/>
                <arg name='properties' direction='out' type='a{sv}'/>
            </method>
            <signal name='PropertiesChanged'>
                <arg name='interface_name' type='s' />
                <arg name='changed_properties' type='a{sv}' />
                <arg name='invalidated_properties' type='as' />
            </signal>
        </interface>
        <interface name='org.mpris.MediaPlayer2'>
            <method name='Raise'>
            </method>
            <method name='Quit'>
            </method>
            <property name='CanQuit' type='b' access='read' />
            <property name='Fullscreen' type='b' access='readwrite' />
            <property name='CanRaise' type='b' access='read' />
            <property name='HasTrackList' type='b' access='read'/>
            <property name='Identity' type='s' access='read'/>
            <property name='DesktopEntry' type='s' access='read'/>
            <property name='SupportedUriSchemes' type='as' access='read'/>
            <property name='SupportedMimeTypes' type='as' access='read'/>
        </interface>
        <interface name='org.mpris.MediaPlayer2.Player'>
            <method name='Next'/>
            <method name='Previous'/>
            <method name='Pause'/>
            <method name='PlayPause'/>
            <method name='Stop'/>
            <method name='Play'/>
            <method name='Seek'>
                <arg direction='in' name='Offset' type='x'/>
            </method>
            <method name='SetPosition'>
                <arg direction='in' name='TrackId' type='o'/>
                <arg direction='in' name='Position' type='x'/>
            </method>
            <method name='OpenUri'>
                <arg direction='in' name='Uri' type='s'/>
            </method>
            <signal name='Seeked'>
                <arg name='Position' type='x'/>
            </signal>
            <property name='PlaybackStatus' type='s' access='read'/>
            <property name='LoopStatus' type='s' access='readwrite'/>
            <property name='Rate' type='d' access='readwrite'/>
            <property name='Shuffle' type='b' access='readwrite'/>
            <property name='Metadata' type='a{sv}' access='read'>
            </property>
            <property name='Position' type='x' access='read'/>
            <property name='MinimumRate' type='d' access='read'/>
            <property name='MaximumRate' type='d' access='read'/>
            <property name='CanGoNext' type='b' access='read'/>
            <property name='CanGoPrevious' type='b' access='read'/>
            <property name='CanPlay' type='b' access='read'/>
            <property name='CanPause' type='b' access='read'/>
            <property name='CanSeek' type='b' access='read'/>
            <property name='CanControl' type='b' access='read'/>
        </interface>
        <interface name='org.mpris.MediaPlayer2.TrackList'>
            <method name='GetTracksMetadata'>
                <arg direction='in' name='TrackIds' type='ao'/>
                <arg direction='out' name='Metadata' type='aa{sv}'>
                </arg>
            </method>
            <method name='AddTrack'>
                <arg direction='in' name='Uri' type='s'/>
                <arg direction='in' name='AfterTrack' type='o'/>
                <arg direction='in' name='SetAsCurrent' type='b'/>
            </method>
            <method name='RemoveTrack'>
                <arg direction='in' name='TrackId' type='o'/>
            </method>
            <method name='GoTo'>
                <arg direction='in' name='TrackId' type='o'/>
            </method>
            <signal name='TrackListReplaced'>
                <arg name='Tracks' type='ao'/>
                <arg name='CurrentTrack' type='o'/>
            </signal>
            <property name='Tracks' type='ao' access='read'/>
            <property name='CanEditTracks' type='b' access='read'/>
        </interface>
        <interface name='org.mpris.MediaPlayer2.Playlists'>
            <method name='ActivatePlaylist'>
                <arg direction="in" name="PlaylistId" type="o" />
            </method>
            <method name='GetPlaylists'>
                <arg direction='in' name='Index' type='u' />
                <arg direction='in' name='MaxCount' type='u' />
                <arg direction='in' name='Order' type='s' />
                <arg direction='in' name='ReverseOrder' type='b' />
                <arg direction='out' name='Playlists' type='a(oss)' />
            </method>
            <property name='PlaylistCount' type='u' access='read' />
            <property name='Orderings' type='as' access='read' />
            <property name='ActivePlaylist' type='(b(oss))' access='read' />
            <signal name='PlaylistChanged'>
                <arg name='Playlist' type='(oss)' />
            </signal>
        </interface>
    </node>
    """

    MEDIA_PLAYER2_IFACE = 'org.mpris.MediaPlayer2'
    MEDIA_PLAYER2_PLAYER_IFACE = 'org.mpris.MediaPlayer2.Player'
    MEDIA_PLAYER2_TRACKLIST_IFACE = 'org.mpris.MediaPlayer2.TrackList'
    MEDIA_PLAYER2_PLAYLISTS_IFACE = 'org.mpris.MediaPlayer2.Playlists'

    _playlist_nb_songs = 10

    def __init__(self, app):
        name = "org.mpris.MediaPlayer2.{}".format(app.props.application_id)
        path = '/org/mpris/MediaPlayer2'
        super().__init__(name, path, app)

        self._app = app
        self._log = app.props.log
        self._player = app.props.player
        self._player.connect(
            'song-changed', self._on_current_song_changed)
        self._player.connect('notify::state', self._on_player_state_changed)
        self._player.connect(
            'notify::repeat-mode', self._on_repeat_mode_changed)
        self._player.connect('seek-finished', self._on_seek_finished)

        self._coremodel = app.props.coremodel
        self._player_model = self._coremodel.props.playlist_sort
        self._player_model_changed_id = None

        self._coremodel.connect(
            "playlist-loaded", self._on_player_playlist_changed)

        self._playlists_model = self._coremodel.props.playlists_sort
        n_items = self._playlists_model.get_n_items()
        self._on_playlists_items_changed(
            self._playlists_model, 0, n_items, n_items)
        self._playlists_model.connect(
            "items-changed", self._on_playlists_items_changed)

        self._recent_playlist = self._coremodel.props.recent_playlist
        self._recent_playlist.connect(
            "items-changed", self._on_recent_playlist_changed)

        self._player_playlist_type = None
        self._path_list = []
        self._metadata_list = []
        self._previous_can_go_next = False
        self._previous_can_go_previous = False
        self._previous_can_play = False
        self._previous_is_shuffled = None
        self._previous_loop_status = ""
        self._previous_mpris_playlist = self._get_active_playlist()
        self._previous_playback_status = "Stopped"
        self._thumbnail_id = 0
        self._current_coresong = None

    def _get_playback_status(self):
        state = self._player.props.state
        if state == Playback.STOPPED:
            return 'Stopped'
        elif state == Playback.PAUSED:
            return 'Paused'
        else:
            return 'Playing'

    def _get_loop_status(self):
        if self._player.props.repeat_mode == RepeatMode.ALL:
            return "Playlist"
        elif self._player.props.repeat_mode == RepeatMode.SONG:
            return "Track"
        else:
            return "None"

    def _get_metadata(self, coresong=None, index=None):
        song_dbus_path = self._get_song_dbus_path(coresong, index)
        if not self._player.props.current_song:
            return {
                'mpris:trackid': GLib.Variant('o', song_dbus_path)
            }

        if not coresong:
            coresong = self._player.props.current_song

        length = coresong.props.duration * 1e6
        user_rating = 1.0 if coresong.props.favorite else 0.0
        artist = coresong.props.artist

        metadata = {
            'mpris:trackid': GLib.Variant('o', song_dbus_path),
            'xesam:url': GLib.Variant('s', coresong.props.url),
            'mpris:length': GLib.Variant('x', length),
            'xesam:useCount': GLib.Variant('i', coresong.props.play_count),
            'xesam:userRating': GLib.Variant('d', user_rating),
            'xesam:title': GLib.Variant('s', coresong.props.title),
            'xesam:album': GLib.Variant('s', coresong.props.album),
            'xesam:artist': GLib.Variant('as', [artist]),
            'xesam:albumArtist': GLib.Variant('as', [artist])
        }

        genre = coresong.props.media.get_genre()
        if genre is not None:
            metadata['xesam:genre'] = GLib.Variant('as', [genre])

        last_played = coresong.props.media.get_last_played()
        if last_played is not None:
            last_played_str = last_played.format("%FT%T%:z")
            metadata['xesam:lastUsed'] = GLib.Variant('s', last_played_str)

        track_nr = coresong.props.track_number
        if track_nr > 0:
            metadata['xesam:trackNumber'] = GLib.Variant('i', track_nr)

        art_url = coresong.props.thumbnail
        if (art_url == "generic"
                and self._current_coresong == coresong):
            self._thumbnail_id = coresong.connect(
                "notify::thumbnail", self._on_thumbnail_changed)
        else:
            metadata['mpris:artUrl'] = GLib.Variant('s', art_url)

        return metadata

    def _on_thumbnail_changed(
            self, coresong: CoreSong, param: GObject.ParamSpecString) -> None:
        """Updates MPRIS metadata when a song's thumbnail changes.

        :param coresong: The song with the new thumbnail.
        :param param_spec: Metadata about the changed property.
        :param user_data: Optional; unused.
        """
        if coresong != self._player.props.current_song:
            return

        properties = {}
        properties["Metadata"] = GLib.Variant(
            "a{sv}", self._get_metadata(coresong))
        self._properties_changed(
            MPRIS.MEDIA_PLAYER2_PLAYER_IFACE, properties, [])

    def _get_song_dbus_path(self, coresong=None, index=None):
        """Convert a Grilo media to a D-Bus path

        The hex encoding is used to remove any possible invalid
        character. Use player index to make the path truly unique in
        case the same song is present multiple times in a playlist.
        If coresong is None, it means that the current song path is
        requested.

        :param CoreSong coresong: The CoreSong object
        :param int index: The media position in the current playlist
        :return: a D-Bus id to uniquely identify the song
        :rtype: str
        """
        if not self._player.props.current_song:
            return "/org/mpris/MediaPlayer2/TrackList/NoTrack"

        if not coresong:
            coresong = self._player.props.current_song
            index = self._player.props.position

        id_hex = coresong.props.grlid.encode('ascii').hex()
        path = "/org/gnome/GnomeMusic/TrackList/{}_{}".format(
            id_hex, index)
        return path

    def _on_recent_playlist_changed(self, model, position, removed, added):
        self._path_list = []
        self._metadata_list = []

        offset = self._recent_playlist.get_offset()
        for position, coresong in enumerate(self._recent_playlist):
            offset_position = position + offset
            self._path_list.append(
                self._get_song_dbus_path(coresong, offset_position))
            self._metadata_list.append(
                self._get_metadata(coresong, offset_position))

        current_song_path = self._get_song_dbus_path()
        self._track_list_replaced(self._path_list, current_song_path)

    def _get_playlist_dbus_path(self, playlist):
        """Convert a playlist to a D-Bus path

        :param Playlist playlist: The playlist object
        :return: a D-Bus id to uniquely identify the playlist
        :rtype: str
        """
        if not playlist:
            return "/"

        # Smart Playlists do not have an id
        pl_id = playlist.props.pl_id or playlist.props.tag_text
        pl_id = pl_id.rsplit(":")[-1].replace("-", "")
        return "/org/gnome/GnomeMusic/Playlist/{}".format(pl_id)

    def _get_mpris_playlist_from_playlist(self, playlist):
        playlist_name = playlist.props.title
        path = self._get_playlist_dbus_path(playlist)
        return (path, playlist_name, "")

    def _get_active_playlist(self):
        """Get Active Maybe_Playlist

        Maybe_Playlist is a structure describing a playlist, or nothing
        according to MPRIS specifications.
        If a playlist is active, return True and its description
        (path, name and icon).
        If no playlist is active, return False and an undefined structure.

        :returns: playlist existence and its structure
        :rtype: tuple
        """
        current_core_object = self._coremodel.props.active_core_object
        if not isinstance(current_core_object, Playlist):
            return (False, ("/", "", ""))

        mpris_playlist = self._get_mpris_playlist_from_playlist(
            current_core_object)
        return (True, mpris_playlist)

    def _on_current_song_changed(self, player: Player) -> None:
        if (self._thumbnail_id != 0
                and self._current_coresong):
            self._current_coresong.disconnect(self._thumbnail_id)
            self._thumbnail_id = 0

        self._current_coresong = player.props.current_song

        # In repeat song mode, no metadata has changed if the
        # player was already started
        if self._player.props.repeat_mode == RepeatMode.SONG:
            self._seeked(0)
            if self._previous_can_play is True:
                return

        if self._player_model_changed_id is None:
            self._player_model_changed_id = self._player_model.connect_after(
                "items-changed", self._on_player_model_changed)

        self._on_player_model_changed(self._player_model, 0, 0, 0)

    def _on_player_model_changed(self, model, pos, removed, added):
        # Do no update the properties if the model has completely changed.
        # These changes will be applied once a new song starts playing.
        if added == model.get_n_items():
            return

        properties = {}
        properties["Metadata"] = GLib.Variant("a{sv}", self._get_metadata())

        has_next = self._player.props.has_next
        if has_next != self._previous_can_go_next:
            properties["CanGoNext"] = GLib.Variant("b", has_next)
            self._previous_can_go_next = has_next

        has_previous = self._player.props.has_previous
        if has_previous != self._previous_can_go_previous:
            properties["CanGoPrevious"] = GLib.Variant("b", has_previous)
            self._previous_can_go_previous = has_previous

        if self._previous_can_play is not True:
            properties["CanPause"] = GLib.Variant("b", True)
            properties["CanPlay"] = GLib.Variant("b", True)
            self._previous_can_play = True

        self._properties_changed(
            MPRIS.MEDIA_PLAYER2_PLAYER_IFACE, properties, [])

    def _on_player_state_changed(self, klass, args):
        playback_status = self._get_playback_status()
        if playback_status == self._previous_playback_status:
            return

        self._previous_playback_status = playback_status
        self._properties_changed(
            MPRIS.MEDIA_PLAYER2_PLAYER_IFACE,
            {'PlaybackStatus': GLib.Variant('s', playback_status), }, [])

    def _on_repeat_mode_changed(self, player, param):
        properties = {}

        is_shuffled = self._player.props.repeat_mode == RepeatMode.SHUFFLE
        if is_shuffled != self._previous_is_shuffled:
            properties["Shuffle"] = GLib.Variant("b", is_shuffled)
            self._previous_is_shuffled = is_shuffled

        loop_status = self._get_loop_status()
        if loop_status != self._previous_loop_status:
            properties["LoopStatus"] = GLib.Variant("s", loop_status)
            self._previous_loop_status = loop_status

        if not properties:
            return
        self._properties_changed(
            MPRIS.MEDIA_PLAYER2_PLAYER_IFACE, properties, [])

    def _on_seek_finished(self, player):
        position_second = self._player.get_position()
        self._seeked(int(position_second * 1e6))

    def _on_player_playlist_changed(self, coremodel, playlist_type):
        self._player_playlist_type = playlist_type

        mpris_playlist = self._get_active_playlist()
        if mpris_playlist == self._previous_mpris_playlist:
            return

        self._previous_mpris_playlist = mpris_playlist
        properties = {
            "ActivePlaylist": GLib.Variant("(b(oss))", mpris_playlist)}
        self._properties_changed(
            MPRIS.MEDIA_PLAYER2_PLAYLISTS_IFACE, properties, [])

    def _on_playlists_items_changed(self, model, position, removed, added):
        if added > 0:
            for i in range(added):
                playlist = model[position + i]
                playlist.connect("notify::title", self._on_playlist_renamed)
                playlist.connect(
                    "notify::active", self._on_player_playlist_changed)

        playlist_count = model.get_n_items()
        properties = {"PlaylistCount": GLib.Variant("u", playlist_count)}
        self._properties_changed(
            MPRIS.MEDIA_PLAYER2_PLAYLISTS_IFACE, properties, [])

    def _on_playlist_renamed(self, playlist, param):
        mpris_playlist = self._get_mpris_playlist_from_playlist(playlist)
        self._dbus_emit_signal('PlaylistChanged', {'Playlist': mpris_playlist})

    def _raise(self):
        """Brings user interface to the front (MPRIS Method)."""
        self._app.do_activate()

    def _quit(self):
        """Causes the media player to stop running (MPRIS Method)."""
        self._app.quit()

    def _next(self):
        """Skips to the next track in the tracklist (MPRIS Method)."""
        self._player.next()

    def _previous(self):
        """Skips to the previous track in the tracklist.

        (MPRIS Method)
        """
        self._player.previous()

    def _pause(self):
        """Pauses playback (MPRIS Method)."""
        self._player.pause()

    def _play_pause(self):
        """Play or Pauses playback (MPRIS Method)."""
        self._player.play_pause()

    def _stop(self):
        """Stop playback (MPRIS Method)."""
        self._player.stop()

    def _play(self):
        """Start or resume playback (MPRIS Method).

        If there is no track to play, this has no effect.
        """
        self._player.play()

    def _seek(self, offset_msecond):
        """Seek forward in the current track (MPRIS Method).

        Seek is relative to the current player position.
        If the value passed in would mean seeking beyond the end of the track,
        acts like a call to Next.

        :param int offset_msecond: number of microseconds
        """
        current_position_second = self._player.get_position()
        new_position_second = current_position_second + offset_msecond / 1e6

        duration_second = self._player.props.duration
        if new_position_second <= duration_second:
            self._player.set_position(new_position_second)
        else:
            self._player.next()

    def _set_position(self, track_id, position_msecond):
        """Set the current track position in microseconds (MPRIS Method)

        :param str track_id: The currently playing track's identifier
        :param int position_msecond: new position in microseconds
        """
        metadata = self._get_metadata()
        current_track_id = metadata["mpris:trackid"].get_string()
        if track_id != current_track_id:
            return
        self._player.set_position(position_msecond / 1e6)

    def _open_uri(self, uri):
        """Opens the Uri given as an argument (MPRIS Method).

        Not implemented.

        :param str uri: Uri of the track to load.
        """
        pass

    def _seeked(self, position_msecond):
        """Indicate that the track position has changed.

        :param int position_msecond: new position in microseconds.
        """
        self._dbus_emit_signal("Seeked", {"Position": position_msecond})

    def _get_tracks_metadata(self, track_paths):
        """Gets all the metadata available for a set of tracks.

        (MPRIS Method)

        :param list track_paths: list of track ids
        :returns: Metadata of the set of tracks given as input.
        :rtype: list
        """
        metadata = []
        for path in track_paths:
            index = self._path_list.index(path)
            metadata.append(self._metadata_list[index])
        return metadata

    def _add_track(self, uri, after_track, set_as_current):
        """Adds a URI in the TrackList. (MPRIS Method).

        This is not implemented (CanEditTracks is set to False).
        """
        pass

    def _remove_track(self, path):
        """Removes an item from the TrackList. (MPRIS Method).

        This is not implemented (CanEditTracks is set to False).
        """
        pass

    def _go_to(self, path):
        """Skip to the specified TrackId (MPRIS Method).

        :param str path: Identifier of the track to skip to
        """
        current_index = self._path_list.index(self._get_song_dbus_path())
        current_coresong = self._player.props.current_song

        goto_index = self._path_list.index(path)
        new_position = self._player.props.position + goto_index - current_index
        new_coresong = self._player_model[new_position]

        self._player.play(new_coresong)
        current_coresong.props.state = SongWidget.State.PLAYED
        new_coresong.props.state = SongWidget.State.PLAYING

    def _track_list_replaced(self, track_paths, current_song_path):
        """Indicate that the entire tracklist has been replaced.

        (MPRIS Method)

        :param list track_paths: the new list of tracks
        :param current_song_path: the id of the current song
        """
        parameters = {
            "Tracks": track_paths,
            "CurrentTrack": current_song_path}
        self._dbus_emit_signal("TrackListReplaced", parameters)

    def _load_player_playlist(self, playlist):
        def _on_playlist_loaded(klass, playlist_type):
            self._player.play()
            self._coremodel.disconnect(loaded_id)

        loaded_id = self._coremodel.connect(
            "playlist-loaded", _on_playlist_loaded)
        self._coremodel.props.active_core_object = playlist

    def _activate_playlist(self, playlist_path):
        """Starts playing the given playlist (MPRIS Method).

        :param str playlist_path: The id of the playlist to activate.
        """
        selected_playlist = None
        for playlist in self._playlists_model:
            if playlist_path == self._get_playlist_dbus_path(playlist):
                selected_playlist = playlist
                break

        if selected_playlist is None:
            return

        def _on_playlist_model_loaded(playlist):
            playlist.disconnect(signal_id)
            self._load_player_playlist(playlist)

        if selected_playlist.props.model.get_n_items() > 0:
            self._load_player_playlist(selected_playlist)
        else:
            signal_id = selected_playlist.connect(
                "playlist-loaded", _on_playlist_model_loaded)

    def _get_playlists(self, index, max_count, order, reverse):
        """Gets a set of playlists (MPRIS Method).

        GNOME Music only handles playlists with the Alphabetical order.

        :param int index: the index of the first playlist to be fetched
        :param int max_count: the maximum number of playlists to fetch.
        :param str order: the ordering that should be used.
        :param bool reverse: whether the order should be reversed.
        """
        if order != 'Alphabetical':
            return []

        mpris_playlists = [self._get_mpris_playlist_from_playlist(playlist)
                           for playlist in self._playlists_model]

        if not reverse:
            return mpris_playlists[index:index + max_count]

        first_index = index - 1
        if first_index < 0:
            first_index = None
        return mpris_playlists[index + max_count - 1:first_index:-1]

    def _get(self, interface_name, property_name):
        # Some clients (for example GSConnect) try to acesss the volume
        # property. This results in a crash at startup.
        # Return nothing to prevent it.
        try:
            return self._get_all(interface_name)[property_name]
        except KeyError:
            msg = "MPRIS does not handle {} property from {} interface".format(
                property_name, interface_name)
            self._log.warning(msg)
            raise ValueError(msg)

    def _get_all(self, interface_name):
        if interface_name == MPRIS.MEDIA_PLAYER2_IFACE:
            application_id = self._app.props.application_id
            return {
                'CanQuit': GLib.Variant('b', True),
                'Fullscreen': GLib.Variant('b', False),
                'CanSetFullscreen': GLib.Variant('b', False),
                'CanRaise': GLib.Variant('b', True),
                'HasTrackList': GLib.Variant('b', True),
                'Identity': GLib.Variant('s', 'Music'),
                'DesktopEntry': GLib.Variant('s', application_id),
                'SupportedUriSchemes': GLib.Variant('as', [
                    'file'
                ]),
                'SupportedMimeTypes': GLib.Variant('as', [
                    'application/ogg',
                    'audio/x-vorbis+ogg',
                    'audio/x-flac',
                    'audio/mpeg'
                ]),
            }
        elif interface_name == MPRIS.MEDIA_PLAYER2_PLAYER_IFACE:
            position_msecond = int(self._player.get_position() * 1e6)
            playback_status = self._get_playback_status()
            is_shuffle = (self._player.props.repeat_mode == RepeatMode.SHUFFLE)
            can_play = (self._player.props.current_song is not None)
            has_previous = (self._player.props.has_previous)
            return {
                'PlaybackStatus': GLib.Variant('s', playback_status),
                'LoopStatus': GLib.Variant('s', self._get_loop_status()),
                'Rate': GLib.Variant('d', 1.0),
                'Shuffle': GLib.Variant('b', is_shuffle),
                'Metadata': GLib.Variant('a{sv}', self._get_metadata()),
                'Position': GLib.Variant('x', position_msecond),
                'MinimumRate': GLib.Variant('d', 1.0),
                'MaximumRate': GLib.Variant('d', 1.0),
                'CanGoNext': GLib.Variant('b', self._player.props.has_next),
                'CanGoPrevious': GLib.Variant('b', has_previous),
                'CanPlay': GLib.Variant('b', can_play),
                'CanPause': GLib.Variant('b', can_play),
                'CanSeek': GLib.Variant('b', True),
                'CanControl': GLib.Variant('b', True),
            }
        elif interface_name == MPRIS.MEDIA_PLAYER2_TRACKLIST_IFACE:
            return {
                'Tracks': GLib.Variant('ao', self._path_list),
                'CanEditTracks': GLib.Variant('b', False)
            }
        elif interface_name == MPRIS.MEDIA_PLAYER2_PLAYLISTS_IFACE:
            playlist_count = self._playlists_model.get_n_items()
            active_playlist = self._get_active_playlist()
            return {
                'PlaylistCount': GLib.Variant('u', playlist_count),
                'Orderings': GLib.Variant('as', ['Alphabetical']),
                'ActivePlaylist': GLib.Variant('(b(oss))', active_playlist),
            }
        elif interface_name == 'org.freedesktop.DBus.Properties':
            return {}
        elif interface_name == 'org.freedesktop.DBus.Introspectable':
            return {}
        else:
            self._log.warning(
                "MPRIS does not implement {} interface".format(interface_name))

    def _set(self, interface_name, property_name, new_value):
        if interface_name == MPRIS.MEDIA_PLAYER2_IFACE:
            if property_name == 'Fullscreen':
                pass
        elif interface_name == MPRIS.MEDIA_PLAYER2_PLAYER_IFACE:
            if property_name in ['Rate', 'Volume']:
                pass
            elif property_name == 'LoopStatus':
                if new_value == 'None':
                    self._player.props.repeat_mode = RepeatMode.NONE
                elif new_value == 'Track':
                    self._player.props.repeat_mode = RepeatMode.SONG
                elif new_value == 'Playlist':
                    self._player.props.repeat_mode = RepeatMode.ALL
            elif property_name == 'Shuffle':
                if new_value:
                    self._player.props.repeat_mode = RepeatMode.SHUFFLE
                else:
                    self._player.props.repeat_mode = RepeatMode.NONE
        else:
            self._log.warning(
                "MPRIS does not implement {} interface".format(interface_name))

    def _properties_changed(self, interface_name, changed_properties,
                            invalidated_properties):
        parameters = {
            'interface_name': interface_name,
            'changed_properties': changed_properties,
            'invalidated_properties': invalidated_properties
        }
        self._dbus_emit_signal('PropertiesChanged', parameters)

    def _introspect(self):
        return self.__doc__



# Copyright 2020 The GNOME Music developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

import inspect
import os

from gi.repository import GLib, GObject


class MusicLogger(GObject.GObject):
    """GLib logging wrapper

    A tiny wrapper around the default GLib logger.

    * Message is for user facing warnings, which ideally should be in
      the application.
    * Warning is for logging non-fatal errors during execution.
    * Debug is for developer use as a way to get more runtime info.
    """

    _DOMAIN = "org.gnome.Music"

    def _log(self, message, level):
        stack = inspect.stack()

        filename = os.path.basename(stack[2][1])
        line = stack[2][2]
        function = stack[2][3]

        if level in [GLib.LogLevelFlags.LEVEL_DEBUG,
                     GLib.LogLevelFlags.LEVEL_INFO,
                     GLib.LogLevelFlags.LEVEL_WARNING]:
            message = "({}, {}, {}) {}".format(
                filename, function, line, message)

        variant_message = GLib.Variant("s", message)
        variant_file = GLib.Variant("s", filename)
        variant_line = GLib.Variant("i", line)
        variant_func = GLib.Variant("s", function)

        variant_dict = GLib.Variant("a{sv}", {
            "MESSAGE": variant_message,
            "CODE_FILE": variant_file,
            "CODE_LINE": variant_line,
            "CODE_FUNC": variant_func
        })

        GLib.log_variant(self._DOMAIN, level, variant_dict)

    def message(self, message):
        """The default user facing message

        Wraps g_message.

        :param string message: Message
        """
        self._log(message, GLib.LogLevelFlags.LEVEL_MESSAGE)

    def warning(self, message):
        """Warning message

        Wraps g_warning.

        :param string message: Warning message
        """
        self._log(message, GLib.LogLevelFlags.LEVEL_WARNING)

    def info(self, message):
        """Informational message

        Wraps g_info.

        :param string message: Informational message
        """
        self._log(message, GLib.LogLevelFlags.LEVEL_INFO)

    def debug(self, message):
        """Debug message

        Wraps g_debug.

        :param string message: Debug message
        """
        self._log(message, GLib.LogLevelFlags.LEVEL_DEBUG)



# Copyright 2024 The GNOME Music developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from __future__ import annotations
from typing import Optional
import asyncio
import os
import typing

from gi.repository import GLib, Gio, GObject

from gnomemusic.gstplayer import Playback
from gnomemusic.musiclogger import MusicLogger

if typing.TYPE_CHECKING:
    from gnomemusic.player import Player


class PauseOnSuspend(GObject.GObject):
    """PauseOnSuspend object

    Contains logic to pause music on system suspend
    and inhibit suspend before pause.
    """

    def __init__(self, player: Player) -> None:
        """Initialize pause on supend handling

        :param Player player: Player object
        """
        super().__init__()

        self._log = MusicLogger()

        self._player = player

        self._conn_signal_id: int
        self._file_descriptor: int = -1
        self._suspend_proxy: Optional[Gio.DBusProxy] = None
        self._previous_state: Playback = self._player.props.state
        self._player.connect("notify::state", self._on_player_state_changed)

        asyncio.create_task(self._init_pause_on_suspend())

    async def _init_pause_on_suspend(self) -> None:
        try:
            self._suspend_proxy = await Gio.DBusProxy.new_for_bus(
                Gio.BusType.SYSTEM,
                Gio.DBusProxyFlags.DO_NOT_LOAD_PROPERTIES,
                None,
                "org.freedesktop.ConsoleKit",
                "/org/freedesktop/ConsoleKit/Manager",
                "org.freedesktop.ConsoleKit.Manager")
        except GLib.Error as error:
            self._log.warning(
                f"Error: Failed to contact logind daemon: {error.message}")

    def _on_player_state_changed(self, player: Player, state: int) -> None:
        if not self._suspend_proxy:
            return

        new_state = self._player.props.state
        if self._previous_state == new_state:
            return

        if (new_state == Playback.PLAYING
                and self._file_descriptor == -1):
            asyncio.create_task(self._take_lock())

        if (self._previous_state == Playback.PLAYING
                and new_state != Playback.LOADING):
            self._release_lock()

        self._previous_state = new_state

    async def _take_lock(self) -> None:
        if not self._suspend_proxy:
            return

        variant = GLib.Variant(
            "(ssss)",
            (
                "sleep",
                "GNOME Music",
                "GNOME Music is pausing",
                "delay"
            )
        )

        try:
            var = await self._suspend_proxy.call_with_unix_fd_list(
                "Inhibit", variant, Gio.DBusCallFlags.NONE, -1)
            self._file_descriptor = var.out_fd_list.get(0)
            self._conn_signal_id = self._suspend_proxy.connect(
                "g-signal", self._pause_playing)
        except GLib.Error as error:
            self._log.warning(
                f"Error: Failed to finish proxy call: {error.message}")

    def _release_lock(self) -> None:
        if not self._suspend_proxy:
            return

        if self._file_descriptor >= 0:
            os.close(self._file_descriptor)
            self._file_descriptor = -1
            self._suspend_proxy.disconnect(self._conn_signal_id)

    def _pause_playing(
            self, proxy: Gio.DBusProxy, sender: Optional[str], signal: str,
            parameters: GLib.Variant) -> None:
        if signal != "PrepareForSleep":
            return

        (going_to_sleep, ) = parameters
        if going_to_sleep is True:
            self._player.pause()
            self._release_lock()
        else:
            asyncio.create_task(self._take_lock())



# Copyright 2024 The GNOME Music developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from __future__ import annotations
import typing

from gi.repository import Adw, GObject, Gtk

from gnomemusic.coreartist import CoreArtist
from gnomemusic.widgets.artistalbumswidget import ArtistAlbumsWidget
if typing.TYPE_CHECKING:
    from gnomemusic.application import Application


@Gtk.Template(resource_path='/org/gnome/Music/ui/ArtistNavigationPage.ui')
class ArtistNavigationPage(Adw.NavigationPage):
    """ArtistNavigationPage
    """

    __gtype_name__ = "ArtistNavigationPage"

    _artist_scrolled_window = Gtk.Template.Child()

    show_artist_label = GObject.Property(type=bool, default=True)

    def __init__(
            self, application: Application,
            coreartist: CoreArtist) -> None:
        """Initialize the AlbumNavigationPage.

        :param GtkApplication application: The application object
        :param CoreArtist coreartist: The artist albums to show
        """
        super().__init__()

        self._application = application
        self._coreartist = coreartist

        self.props.title = self._coreartist.props.artist

        artist_albums = ArtistAlbumsWidget(self._application)
        artist_albums.props.coreartist = self._coreartist

        self._artist_scrolled_window.props.child = artist_albums



# Copyright (c) 2013 Arnel A. Borja <kyoushuu@yahoo.com>
# Copyright (c) 2013 Vadim Rutkovsky <vrutkovs@redhat.com>
# Copyright (c) 2013 Lubosz Sarnecki <lubosz@gmail.com>
# Copyright (c) 2013 Guillaume Quintard <guillaume.quintard@gmail.com>
# Copyright (c) 2013 Felipe Borges <felipe10borges@gmail.com>
# Copyright (c) 2013 Eslam Mostafa <cseslam@gmail.com>
# Copyright (c) 2013 Shivani Poddar <shivani.poddar92@gmail.com>
# Copyright (c) 2013 Sai Suman Prayaga <suman.sai14@gmail.com>
# Copyright (c) 2013 Seif Lotfy <seif@lotfy.com>
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

import asyncio
from typing import Optional
from gettext import gettext as _

from gi.events import GLibEventLoopPolicy
from gi.repository import Adw, Gtk, Gio, GLib, Gdk, GObject

from gnomemusic.about import show_about
from gnomemusic.coregrilo import CoreGrilo
from gnomemusic.coremodel import CoreModel
from gnomemusic.inhibitsuspend import InhibitSuspend
from gnomemusic.mpris import MPRIS
from gnomemusic.musiclogger import MusicLogger
from gnomemusic.notificationmanager import NotificationManager
from gnomemusic.pauseonsuspend import PauseOnSuspend
from gnomemusic.player import Player
from gnomemusic.search import Search
from gnomemusic.widgets.preferencesdialog import PreferencesDialog
from gnomemusic.window import Window


class Application(Adw.Application):

    def __init__(self, application_id, version):
        super().__init__(
            application_id=application_id,
            flags=Gio.ApplicationFlags.FLAGS_NONE)
        self.props.resource_base_path = "/org/gnome/Music"
        GLib.set_application_name(_("Music"))
        GLib.set_prgname(application_id)
        GLib.setenv("PULSE_PROP_media.role", "music", True)

        asyncio.set_event_loop_policy(GLibEventLoopPolicy())

        self._version = version
        self._window = None

        self._log = MusicLogger()
        self._search = Search()

        self._notificationmanager = NotificationManager(self)
        self._coremodel = CoreModel(self)
        # Order is important: CoreGrilo initializes the Grilo sources,
        # which in turn uses CoreModel extensively.
        self._coregrilo = CoreGrilo(self)

        self._settings = Gio.Settings.new('org.gnome.Music')
        self._player = Player(self)

        InhibitSuspend(self)

    @GObject.Property(
        type=CoreGrilo, default=None, flags=GObject.ParamFlags.READABLE)
    def coregrilo(self):
        """Get application-wide CoreGrilo instance.

        :returns: The grilo wrapper
        :rtype: CoreGrilo
        """
        return self._coregrilo

    @GObject.Property(
        type=MusicLogger, default=None, flags=GObject.ParamFlags.READABLE)
    def log(self):
        """Get application-wide logging facility.

        :returns: the logger
        :rtype: MusicLogger
        """
        return self._log

    @GObject.Property(
        type=Player, default=None, flags=GObject.ParamFlags.READABLE)
    def player(self):
        """Get application-wide music player.

        :returns: the player
        :rtype: Player
        """
        return self._player

    @GObject.Property(
        type=Gio.Settings, flags=GObject.ParamFlags.READABLE)
    def settings(self):
        """Get application-wide settings.

        :returns: settings
        :rtype: Gio.settings
        """
        return self._settings

    @GObject.Property(
        type=CoreModel, flags=GObject.ParamFlags.READABLE)
    def coremodel(self):
        """Get class providing all listmodels.

        :returns: List model provider class
        :rtype: CoreModel
        """
        return self._coremodel

    @GObject.Property(type=Window, flags=GObject.ParamFlags.READABLE)
    def window(self):
        """Get main window.

        :returns: Main window.
        :rtype: Window
        """
        return self._window

    @GObject.Property(
        type=Search, flags=GObject.ParamFlags.READABLE)
    def search(self):
        """Get class providing search logic.

        :returns: List model provider class
        :rtype: Search
        """
        return self._search

    @GObject.Property(
        type=NotificationManager, flags=GObject.ParamFlags.READABLE)
    def notificationmanager(self):
        """Get notification manager

        :returns: notification manager
        :rtype: NotificationManager
        """
        return self._notificationmanager

    def _set_actions(self):
        action_entries = [
            ("about", self._about, None),
            ("preferences", self._preferences_dialog,
                ("app.preferences", ["<Ctrl>comma"])),
            ("help", self._help, ("app.help", ["F1"])),
            ("quit", self._quit, ("app.quit", ["<Ctrl>Q"]))
        ]

        for action, callback, accel in action_entries:
            simple_action = Gio.SimpleAction.new(action, None)
            simple_action.connect('activate', callback)
            self.add_action(simple_action)
            if accel is not None:
                self.set_accels_for_action(*accel)

    def _help(self, action: Gio.Action, param: Optional[GLib.Variant]) -> None:

        def show_uri_cb(parent: Gtk.Window, result: Gio.AsyncResult) -> None:
            try:
                Gtk.show_uri_full_finish(parent, result)
            except GLib.Error:
                self._log.message("Help handler not available.")

        Gtk.show_uri_full(
            self._window, "help:gnome-music", Gdk.CURRENT_TIME, None,
            show_uri_cb)

    def _preferences_dialog(
            self, action: Gio.SimpleAction,
            param_type: GLib.VariantType) -> None:
        if self._window.props.visible_dialog:
            return

        pref_dialog = PreferencesDialog(self)
        pref_dialog.present(self._window)

    def _about(self, action, param):
        show_about(self.props.application_id, self._version, self._window)

    def do_startup(self):
        Adw.Application.do_startup(self)
        Adw.StyleManager.get_default().set_color_scheme(
            Adw.ColorScheme.PREFER_LIGHT)
        self._set_actions()

    def _quit(self, action=None, param=None):
        self._window.destroy()

    def do_activate(self):
        if not self._window:
            self._window = Window(self)
            self.notify("window")
            self._window.set_default_icon_name(self.props.application_id)
            if self.props.application_id == "org.gnome.Music.Devel":
                self._window.get_style_context().add_class('devel')
            MPRIS(self)

        PauseOnSuspend(self._player)

        self._window.present()



# Copyright 2024 The GNOME Music developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from __future__ import annotations
import typing

from gi.repository import Adw, GObject, Gtk

from gnomemusic.corealbum import CoreAlbum
from gnomemusic.widgets.albumwidget import AlbumWidget
if typing.TYPE_CHECKING:
    from gnomemusic.application import Application


@Gtk.Template(resource_path='/org/gnome/Music/ui/AlbumNavigationPage.ui')
class AlbumNavigationPage(Adw.NavigationPage):
    """AlbumNavigationPage
    """

    __gtype_name__ = "AlbumNavigationPage"

    _album_scrolled_window = Gtk.Template.Child()
    _page_title = Gtk.Template.Child()

    show_artist_label = GObject.Property(type=bool, default=True)

    def __init__(self, application: Application, corealbum: CoreAlbum) -> None:
        """Initialize the AlbumNavigationPage.

        :param GtkApplication application: The application object
        :param CoreAlbum corealbum: The CoreAlbum to show
        """
        super().__init__()

        self._application = application
        self._corealbum = corealbum

        self._page_title.props.title = self._corealbum.props.title
        self._page_title.props.subtitle = self._corealbum.props.artist

        album_widget = AlbumWidget(self._application)
        album_widget.props.corealbum = self._corealbum

        self._album_scrolled_window.props.child = album_widget



# Copyright (c) 2016 The GNOME Music Developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

from __future__ import annotations
from enum import IntEnum
from typing import Optional
import typing

from gi.repository import Adw, GObject, Gtk
from gettext import gettext as _

from gnomemusic.search import Search
from gnomemusic.utils import ArtSize
from gnomemusic.widgets.albumcover import AlbumCover
from gnomemusic.widgets.albumnavigationpage import AlbumNavigationPage
from gnomemusic.widgets.albumssearchnavigationpage import (
    AlbumsSearchNavigationPage)
from gnomemusic.widgets.artistnavigationpage import ArtistNavigationPage
from gnomemusic.widgets.artistsearchtile import ArtistSearchTile
from gnomemusic.widgets.artistssearchnavigationpage import (
    ArtistsSearchNavigationPage)
from gnomemusic.widgets.searchheaderbar import SearchHeaderBar
from gnomemusic.widgets.songwidget import SongWidget
from gnomemusic.widgets.songwidgetmenu import SongWidgetMenu
if typing.TYPE_CHECKING:
    from gnomemusic.application import Application
    from gnomemusic.corealbum import CoreAlbum
    from gnomemusic.coresong import CoreSong


@Gtk.Template(resource_path="/org/gnome/Music/ui/SearchView.ui")
class SearchView(Adw.NavigationPage):
    """Gridlike view of search results.

    Three sections: artists, albums, songs.
    """

    __gtype_name__ = "SearchView"

    class State(IntEnum):
        """The different states of SearchView
        """
        MAIN = 0
        ALL_ALBUMS = 1
        ALL_ARTISTS = 2
        ALBUM = 3
        ARTIST = 4

    search_mode_active = GObject.Property(type=bool, default=False)
    search_state = GObject.Property(type=int, default=Search.State.NONE)
    search_text = GObject.Property(type=str)

    _album_header = Gtk.Template.Child()
    _album_flowbox = Gtk.Template.Child()
    _artist_header = Gtk.Template.Child()
    _artist_flowbox = Gtk.Template.Child()
    _search_results = Gtk.Template.Child()
    _search_toolbar_view = Gtk.Template.Child()
    _songs_header = Gtk.Template.Child()
    _songs_listbox = Gtk.Template.Child()
    _stack = Gtk.Template.Child()
    _status_page = Gtk.Template.Child()
    _view_all_albums = Gtk.Template.Child()
    _view_all_artists = Gtk.Template.Child()

    def __init__(self, application: Application) -> None:
        """Initialize SearchView

        :param GtkApplication application: The Application object
        """
        super().__init__()

        self.props.name = "search"

        self._application = application
        self._coremodel = application.props.coremodel
        self._model = self._coremodel.props.songs_search
        self._player = self._application.props.player
        self._window = application.props.window
        self._navigation_view = self._window.props.navigation_view
        self._headerbar = self._window.props.headerbar

        self._search_headerbar = SearchHeaderBar(self._application)
        self._search_toolbar_view.add_top_bar(self._search_headerbar)

        self.bind_property(
            "search-text", self._search_headerbar, "search-text")

        self.bind_property(
            "search-state", self._search_headerbar, "search-state",
            GObject.BindingFlags.BIDIRECTIONAL
            | GObject.BindingFlags.SYNC_CREATE)
        self.connect(
            "notify::search-state", self._on_search_state_changed)

        self._navigation_view.connect(
            "replaced", self._on_navigation_view_replaced)

        self._album_model = self._coremodel.props.albums_search
        self._artist_model = self._coremodel.props.artists_search

        self._album_slice = Gtk.SliceListModel.new(self._album_model, 0, 8)
        self._artist_slice = Gtk.SliceListModel.new(self._artist_model, 0, 6)

        self._model.connect_after(
            "items-changed", self._on_model_items_changed)
        self._songs_listbox.bind_model(self._model, self._create_song_widget)
        self._on_model_items_changed(self._model, 0, 0, 0)

        self.bind_property(
            "search_mode_active", self._search_headerbar, "search_mode_active",
            GObject.BindingFlags.BIDIRECTIONAL
            | GObject.BindingFlags.SYNC_CREATE)

        self._album_slice.connect_after(
            "items-changed", self._on_album_model_items_changed)
        self._album_flowbox.bind_model(
            self._album_slice, self._create_album_cover)
        self._application.props.window.connect(
            "notify::default-width", self._on_window_width_change)

        self._artist_slice.connect_after(
            "items-changed", self._on_artist_model_items_changed)
        self._artist_flowbox.bind_model(
            self._artist_slice, self._create_artist_widget)

        self._scrolled_artist_window: Optional[Gtk.ScrolledWindow] = None

        self._set_empty_status_page()
        self._stack.props.visible_child_name = "status"

    def _create_song_widget(self, coresong: CoreSong) -> Gtk.ListBoxRow:
        song_widget = SongWidget(coresong, False, True)
        song_widget.props.show_song_number = False
        song_widget.props.menu = SongWidgetMenu(
            self._application, song_widget, coresong)

        return song_widget

    def _create_album_cover(self, corealbum: CoreAlbum) -> AlbumCover:
        album_cover = AlbumCover(corealbum)

        return album_cover

    def _create_artist_widget(self, coreartist):
        artist_tile = ArtistSearchTile(coreartist)

        return artist_tile

    def _on_album_model_items_changed(self, model, position, removed, added):
        items_found = model.get_n_items() > 0
        self._album_header.props.visible = items_found
        self._album_flowbox.props.visible = items_found
        self._check_visibility()

        nr_albums = self._album_model.get_n_items()
        self._view_all_albums.props.visible = (nr_albums > model.get_n_items())

    def _on_artist_model_items_changed(self, model, position, removed, added):
        items_found = model.get_n_items() > 0
        self._artist_header.props.visible = items_found
        self._artist_flowbox.props.visible = items_found
        self._check_visibility()

        nr_artists = self._artist_model.get_n_items()
        self._view_all_artists.props.visible = (
            nr_artists > model.get_n_items())

    def _on_model_items_changed(self, model, position, removed, added):
        items_found = model.get_n_items() > 0
        self._songs_header.props.visible = items_found
        self._songs_listbox.props.visible = items_found
        self._check_visibility()

    def _on_navigation_view_replaced(self, view: Adw.NavigationView) -> None:
        if view.props.visible_page.props.tag != self.props.tag:
            self.props.search_state = Search.State.NONE

    def _on_search_state_changed(
            self, searchview: SearchView, state: GObject.GParamInt) -> None:
        state = self.props.search_state

        if state == Search.State.RESULT:
            self._stack.props.visible_child_name = "main"
        elif state == Search.State.NONE:
            self._set_empty_status_page()
            self._stack.props.visible_child_name = "status"
        elif state == Search.State.NO_RESULT:
            self._set_no_result_status_page()
            self._stack.props.visible_child_name = "status"

    def _set_empty_status_page(self) -> None:
        self._status_page.props.title = _("No Search Started")
        self._status_page.props.description = _(
            "Use the searchbar to start searching for "
            "albums, artists or songs")

    def _set_no_result_status_page(self) -> None:
        self._status_page.props.title = _("No Results Found")
        self._status_page.props.description = _(
            "Try a different search")

    def _check_visibility(self):
        if not self.props.search_mode_active:
            return

        items_found = (self._model.get_n_items() > 0
                       or self._artist_model.get_n_items() > 0
                       or self._album_model.get_n_items() > 0)
        if items_found:
            self.props.search_state = Search.State.RESULT
        elif (not items_found
              and len(self.props.search_text) > 0):
            self.props.search_state = Search.State.NO_RESULT

    @Gtk.Template.Callback()
    def _song_activated(
            self, list_box: Gtk.ListBox, song_widget: SongWidget) -> bool:
        coresong = song_widget.props.coresong

        self._coremodel.props.active_core_object = coresong
        self._player.play(coresong)

        return True

    def _on_window_width_change(self, widget, value):
        allocation = self._album_flowbox.get_allocation()
        # FIXME: Just a bit of guesswork here.
        padding = 32
        items_per_row = allocation.width // (ArtSize.MEDIUM.width + padding)

        self._album_slice.props.size = 2 * items_per_row
        self._artist_slice.props.size = items_per_row

    @Gtk.Template.Callback()
    def _on_album_activated(self, widget, child, user_data=None):
        corealbum = child.props.corealbum
        album_page = AlbumNavigationPage(self._application, corealbum)
        self._navigation_view.push(album_page)

    @Gtk.Template.Callback()
    def _on_artist_activated(self, widget, child, user_data=None):
        coreartist = child.props.coreartist
        artist_page = ArtistNavigationPage(self._application, coreartist)
        self._navigation_view.push(artist_page)

    @Gtk.Template.Callback()
    def _on_all_artists_clicked(self, widget, user_data=None):
        all_artists_page = ArtistsSearchNavigationPage(
            self._application, self._artist_model)
        self._navigation_view.push(all_artists_page)

    @Gtk.Template.Callback()
    def _on_all_albums_clicked(self, widget, user_data=None):
        all_albums_page = AlbumsSearchNavigationPage(
            self._application, self._album_model)
        self._navigation_view.push(all_albums_page)



# Copyright 2018 The GNOME Music developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

from gi.repository import GObject, Gtk


@Gtk.Template(resource_path='/org/gnome/Music/ui/TwoLineTip.ui')
class TwoLineTip(Gtk.Box):
    """Tooltip with two lines of text

    A bolded title and a regular subtitle.
    """

    __gtype_name__ = 'TwoLineTip'

    _title_label = Gtk.Template.Child()
    _subtitle_label = Gtk.Template.Child()

    title = GObject.Property(type=str)
    subtitle = GObject.Property(type=str)
    subtitle_visible = GObject.Property(type=bool, default=True)

    def __init__(self):
        super().__init__()

        self.bind_property('title', self._title_label, 'label')
        self.bind_property('subtitle', self._subtitle_label, 'label')
        self.bind_property('subtitle-visible', self._subtitle_label, 'visible')



# Copyright 2020 The GNOME Music Developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

from gi.repository import GLib, GObject


class NotificationManager(GObject.Object):
    """Managing wrapper around the notification widgets
    """

    _pulse_id = 0
    _loading_counter = 0

    def __init__(self, application):
        """Initialize the notification manager

       :param Application application: The Application instance
       """
        super().__init__()

        self._application = application
        self._window = application.props.window

        if self._window is None:
            application.connect(
                "notify::window", self._on_window_changed)

    def _on_window_changed(self, klass, value):
        self._window = self._application.props.window
        if self._loading_counter > 0:
            self._window.loading_visible(True)

    def push_loading(self):
        """Push a loading notifcation."""
        self._loading_counter += 1

        if (self._pulse_id == 0
                and self._window):
            self._window.loading_visible(True)
            self._pulse_id = GLib.timeout_add(100, self._window.loading_pulse)

    def pop_loading(self):
        self._loading_counter -= 1

        if (self._loading_counter == 0
                and self._pulse_id != 0):
            GLib.Source.remove(self._pulse_id)
            self._pulse_id = 0
            if self._window:
                self._window.loading_visible(False)



# Copyright 2022 The GNOME Music developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

from __future__ import annotations
from typing import Optional, Union
import typing

import gi
gi.require_versions({"Gdk": "4.0", "Gtk": "4.0", "Gsk": "4.0"})
from gi.repository import Adw, Gsk, Gtk, GObject, Graphene, Gdk

from gnomemusic.texturecache import TextureCache
from gnomemusic.utils import ArtSize, DefaultIconType
if typing.TYPE_CHECKING:
    from gnomemusic.corealbum import CoreAlbum
    from gnomemusic.coreartist import CoreArtist
    from gnomemusic.coresong import CoreSong

if typing.TYPE_CHECKING:
    CoreObject = Union[CoreAlbum, CoreArtist, CoreSong]


class CoverPaintable(GObject.GObject, Gdk.Paintable):
    """An album/artist cover or placeholder

    Provides the full looks. Rounded corners for albums and round for
    artists.
    """

    __gtype_name__ = "CoverPaintable"

    def __init__(
            self, widget: Gtk.Widget, art_size: ArtSize,
            icon_type: DefaultIconType) -> None:
        """Initiliaze CoverPaintable

        :param Gtk.Widget widget: Widget using the cover
        :param ArtSize art_size: Size of the cover
        :param DefaultIconType icon_type: Type of cover
        """
        super().__init__()

        self._art_size = art_size
        self._coreobject: Optional[CoreObject] = None
        self._icon_theme = Gtk.IconTheme.new().get_for_display(
            widget.get_display())
        self._icon_type = icon_type
        self._style_manager = Adw.StyleManager.get_default()
        self._texture = None
        self._texture_cache = TextureCache()
        self._thumbnail_id = 0
        self._widget = widget

        self._style_manager.connect("notify::dark", self._on_dark_changed)

    def do_snapshot(self, snapshot: Gtk.Snapshot, w: float, h: float) -> None:
        if self._texture is not None:
            self._snapshot_texture(self._texture, snapshot, w, h)
        else:
            self._snapshot_fallback_icon(snapshot, w, h)

    def _snapshot_texture(
            self, texture: Gdk.Texture, snapshot: Gtk.Snapshot, w: float,
            h: float) -> None:
        w_s = w
        h_s = h
        ratio = texture.get_height() / texture.get_width()
        # Scale down the image according to the biggest axis
        if ratio > 1:
            w = w / ratio
        else:
            h = h * ratio

        scale_factor = self._widget.props.scale_factor

        snapshot.save()
        snapshot.scale(1.0 / scale_factor, 1.0 / scale_factor)

        rect = Graphene.Rect().init((w_s - w) / 2, (h_s - h) / 2, w, h)
        rect = rect.scale(scale_factor, scale_factor)
        rounded_rect = Gsk.RoundedRect()
        rounded_rect.init_from_rect(rect, self._radius() * scale_factor)
        snapshot.push_rounded_clip(rounded_rect)

        snapshot.append_scaled_texture(
            texture, Gsk.ScalingFilter.TRILINEAR, rect)

        snapshot.pop()
        snapshot.restore()

    def _snapshot_fallback_icon(
            self, snapshot: Gtk.Snapshot, w: float, h: float) -> None:
        rounded_rect = Gsk.RoundedRect()
        rounded_rect.init_from_rect(
            Graphene.Rect().init(0, 0, w, h), self._radius())
        snapshot.push_rounded_clip(rounded_rect)

        i_s = 1 / 3  # Icon scale
        icon_pt = self._icon_theme.lookup_icon(
            self._icon_type.value, None, w * i_s,
            self._widget.props.scale_factor, 0, 0)

        bg_color = Gdk.RGBA()
        bg_color.parse("rgba(95%, 95%, 95%, 1)")
        if self._style_manager.props.dark:
            bg_color.parse("rgba(30%, 30%, 30%, 1)")

        snapshot.append_color(bg_color, Graphene.Rect().init(0, 0, w, h))
        snapshot.translate(
            Graphene.Point().init(
                (w / 2) - (w * (i_s / 2)), (h / 2) - (h * (i_s / 2))))
        snapshot.push_opacity(0.7)
        icon_pt.snapshot(snapshot, w * i_s, h * i_s)
        snapshot.pop()

        snapshot.pop()

    def _radius(self) -> float:
        if self._icon_type == DefaultIconType.ARTIST:
            return 90.0
        elif self._art_size == ArtSize.SMALL:
            return 4.5
        else:
            return 9.0

    def _on_dark_changed(
            self, style_manager: Adw.StyleManager,
            pspec: GObject.ParamSpecBoolean) -> None:
        if self._texture is not None:
            return

        self.invalidate_contents()

    @GObject.Property(type=object, default=None)
    def coreobject(self) -> Optional[CoreObject]:
        """Get the current core object in use

        :returns: The corrent coreobject
        :rtype: Union[CoreAlbum, CoreArtist, CoreSong] or None
        """
        return self._coreobject

    @coreobject.setter  # type: ignore
    def coreobject(self, coreobject: CoreObject) -> None:
        """Update the coreobject used for CoverPaintable

        :param Union[CoreAlbum, CoreArtist, CoreSong] coreobject:
            The coreobject to set
        """
        if coreobject is self._coreobject:
            return

        self._texture_cache.clear_pending_lookup_callback()

        if self._texture:
            self._texture = None
            self.invalidate_contents()

        if self._thumbnail_id != 0:
            self._coreobject.disconnect(self._thumbnail_id)
            self._thumbnail_id = 0

        self._coreobject = coreobject
        self._thumbnail_id = self._coreobject.connect(
            "notify::thumbnail", self._on_thumbnail_changed)

        if self._coreobject.props.thumbnail is not None:
            self._on_thumbnail_changed(self._coreobject, None)

    def _on_thumbnail_changed(
            self, coreobject: CoreObject,
            uri: GObject.ParamSpecString) -> None:
        thumbnail_uri = coreobject.props.thumbnail

        if thumbnail_uri == "generic":
            self._texture = None
            self.invalidate_contents()
            return

        self._texture_cache.connect("texture", self._on_texture_cache)
        self._texture_cache.lookup(thumbnail_uri)

    def _on_texture_cache(
            self, texture_cache: TextureCache, texture: Gdk.Texture) -> None:
        if texture == self._texture:
            return

        self._texture = texture
        self.invalidate_contents()

    @GObject.Property(type=object, flags=GObject.ParamFlags.READWRITE)
    def icon_type(self) -> DefaultIconType:
        """Icon type of the cover

        :returns: The type of the default icon
        :rtype: DefaultIconType
        """
        return self._icon_type

    @icon_type.setter  # type: ignore
    def icon_type(self, value: DefaultIconType) -> None:
        """Set the cover icon type

        :param DefaultIconType value: The default icon type for the
            cover
        """
        self._icon_type = value

        self.invalidate_contents()

    def do_get_flags(self) -> Gdk.PaintableFlags:
        return Gdk.PaintableFlags.SIZE

    def do_get_intrinsic_height(self) -> int:
        return self._art_size.height

    def do_get_intrinsic_width(self) -> int:
        return self._art_size.width



# Copyright 2024 The GNOME Music developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from __future__ import annotations
from collections.abc import Callable
from typing import Optional
import asyncio

import gi
gi.require_versions({"Grl": "0.3"})
from gi.repository import Grl, Gtk, GLib, GObject

from gnomemusic.grilowrappers.playlist import Playlist
from gnomemusic.grilowrappers.smartplaylist import (
    AllSongs, InsufficientTagged, Favorites, RecentlyAdded, RecentlyPlayed,
    NeverPlayed, MostPlayed)
import gnomemusic.utils as utils


class GrlTrackerPlaylists(GObject.GObject):

    __gtype_name__ = "GrlTrackerPlaylists"

    def __init__(self, source, application, tracker_wrapper, songs_hash):
        """Initialize GrlTrackerPlaylists.

        :param Grl.TrackerSource source: The Tracker source to wrap
        :param Application application: Application instance
        :param TrackerWrapper tracker_wrapper: The TrackerWrapper
                                               instance
        :param dict songs_hash: The songs hash table
        """
        super().__init__()

        self._application = application
        self._coremodel = application.props.coremodel
        self._log = application.props.log
        self._source = source
        self._model = self._coremodel.props.playlists
        self._model_filter = self._coremodel.props.playlists_filter
        self._user_model_filter = self._coremodel.props.user_playlists_filter
        self._pls_todelete = []
        self._songs_hash = songs_hash
        self._tracker = tracker_wrapper.props.local_db
        self._tracker_wrapper = tracker_wrapper
        self._notificationmanager = application.props.notificationmanager
        self._window = application.props.window

        user_playlists_filter = Gtk.CustomFilter()
        user_playlists_filter.set_filter_func(self._user_playlists_filter)
        self._user_model_filter.set_filter(user_playlists_filter)

        self._fast_options = Grl.OperationOptions()
        self._fast_options.set_resolution_flags(
            Grl.ResolutionFlags.FAST_ONLY | Grl.ResolutionFlags.IDLE_RELAY)

        self._pl_create_stmt = self._tracker.load_statement_from_gresource(
            "/org/gnome/Music/queries/playlist_create.rq")
        self._pl_delete_stmt = self._tracker.load_statement_from_gresource(
            "/org/gnome/Music/queries/playlist_delete.rq")
        self._pl_query_stmt = self._tracker.load_statement_from_gresource(
            "/org/gnome/Music/queries/playlist_query_playlist.rq")
        self._pl_query_all_stmt = self._tracker.load_statement_from_gresource(
            "/org/gnome/Music/queries/playlist_query_all.rq")

        asyncio.create_task(self._initial_playlists_fill())

    async def _initial_playlists_fill(self):
        args = {
            "source": self._source,
            "application": self._application,
            "tracker_wrapper": self._tracker_wrapper,
            "songs_hash": self._songs_hash
        }

        smart_playlists = {
            "MostPlayed": MostPlayed(**args),
            "NeverPlayed": NeverPlayed(**args),
            "RecentlyPlayed": RecentlyPlayed(**args),
            "RecentlyAdded": RecentlyAdded(**args),
            "Favorites": Favorites(**args),
            "InsufficientTagged": InsufficientTagged(**args),
            "AllSongs": AllSongs(**args),
        }

        for playlist in smart_playlists.values():
            self._model.append(playlist)

        try:
            cursor = await self._pl_query_all_stmt.execute_async()
        except GLib.Error as error:
            self._log.warning(
                f"Playlist query statement: {error.domain}, {error.message}")
            return

        try:
            has_next = await cursor.next_async()
        except GLib.Error as error:
            cursor.close()
            self._log.warning(f"Cursor iteration error: {error.message}")
            return
        while has_next:
            media = utils.create_grilo_media_from_cursor(
                cursor, Grl.MediaType.CONTAINER)
            self._add_user_playlist(media)

            try:
                has_next = await cursor.next_async()
            except GLib.Error as error:
                cursor.close()
                self._log.warning(f"Cursor iteration error: {error.message}")
        else:
            cursor.close()

    def _add_user_playlist(
            self, media: Grl.Media,
            callback: Optional[Callable] = None) -> None:
        playlist = Playlist(
            media=media, source=self._source, application=self._application,
            tracker_wrapper=self._tracker_wrapper, songs_hash=self._songs_hash)
        self._model.append(playlist)

        if callback is not None:
            callback(playlist)

    def _playlists_filter(self, playlist):
        return playlist not in self._pls_todelete

    def _user_playlists_filter(self, playlist):
        return (playlist not in self._pls_todelete
                and playlist.props.is_smart is False)

    def stage_playlist_deletion(self, playlist):
        """Adds playlist to the list of playlists to delete

        :param Playlist playlist: playlist
        """
        self._pls_todelete.append(playlist)
        playlists_filter = Gtk.CustomFilter()
        playlists_filter.set_filter_func(self._playlists_filter)
        self._model_filter.set_filter(playlists_filter)

    def finish_playlist_deletion(
            self, playlist: Playlist, deleted: bool) -> None:
        """Removes playlist from the list of playlists to delete

        :param Playlist playlist: playlist
        :param bool deleted: indicates if the playlist has been deleted
        """
        asyncio.create_task(self._finish_playlist_deletion(
            playlist, deleted))

    async def _finish_playlist_deletion(
            self, playlist: Playlist, deleted: bool) -> None:
        playlists_filter = Gtk.CustomFilter()
        playlists_filter.set_filter_func(self._playlists_filter)
        user_playlists_filter = Gtk.CustomFilter()
        user_playlists_filter.set_filter_func(self._playlists_filter)

        self._pls_todelete.remove(playlist)
        if deleted is False:
            self._model_filter.set_filter(playlists_filter)
            self._user_model_filter.set_filter(user_playlists_filter)
            return

        self._notificationmanager.push_loading()

        self._pl_delete_stmt.bind_string("playlist", playlist.props.pl_id)
        try:
            await self._pl_delete_stmt.update_async()
        except GLib.Error as error:
            self._log.warning(
                f"Failed to delete playlist {playlist.props.title}:"
                f"{error.domain}, {error.message}")
        else:
            for idx, playlist_model in enumerate(self._model):
                if playlist_model is playlist:
                    self._model.remove(idx)
                    break

        self._model_filter.set_filter(playlists_filter)

        self._notificationmanager.pop_loading()

    def create_playlist(self, playlist_title: str, callback: Callable) -> None:
        """Creates a new user playlist.

        :param str playlist_title: playlist title
        :param callback: function to perform once the playlist is created
        """
        asyncio.create_task(self._create_playlist(playlist_title, callback))

    async def _create_playlist(
            self, playlist_title: str, callback: Callable) -> None:
        pl_urn = f"urn:gnomemusic:playlist:{playlist_title}"
        self._pl_create_stmt.bind_string("title", playlist_title)
        self._pl_create_stmt.bind_string("playlist", pl_urn)
        try:
            await self._pl_create_stmt.update_async()
        except GLib.Error as error:
            self._log.warning(
                f"Unable to create playlist {playlist_title}:"
                f" {error.domain}, {error.message}")
            if callback is not None:
                callback(None)
            return

        self._pl_query_stmt.bind_string("playlist", pl_urn)
        try:
            cursor = await self._pl_query_stmt.execute_async()
        except GLib.Error as error:
            cursor.close()
            self._log.warning(
                f"Failed playlist query for {pl_urn}: {error.domain},"
                f" {error.message}")
            return

        try:
            has_next = await cursor.next_async()
        except GLib.Error as error:
            cursor.close()
            self._log.warning(
                f"Cursor iteration error: {error.domain}, {error.message}")
            return
        while has_next:
            media = utils.create_grilo_media_from_cursor(
                cursor, Grl.MediaType.CONTAINER)
            self._add_user_playlist(media, callback)

            try:
                has_next = await cursor.next_async()
            except GLib.Error as error:
                cursor.close()
                self._log.warning(
                    f"Cursor iteration error: {error.domain}, {error.message}")
        else:
            cursor.close()

    def check_smart_playlist_change(self):
        """Check if smart playlists need to be updated.

        A smart playlist needs to be updated in two cases:
        * it is being played (active_playlist)
        * it is visible in PlaylistsView
        """
        active_core_object = self._coremodel.props.active_core_object
        if (isinstance(active_core_object, Playlist)
                and active_core_object.props.is_smart):
            active_core_object.update()
        else:
            self._coremodel.emit("smart-playlist-change")



# Copyright (c) 2021 The GNOME Music Developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

from __future__ import annotations
import typing

from gi.repository import Gdk, Gio, GObject, Gtk

from gnomemusic.widgets.songwidget import SongWidget
from gnomemusic.widgets.songwidgetmenu import SongWidgetMenu
if typing.TYPE_CHECKING:
    from gnomemusic.application import Application
    from gnomemusic.corealbum import CoreAlbum
    from gnomemusic.coredisc import CoreDisc


@Gtk.Template(resource_path='/org/gnome/Music/ui/DiscBox.ui')
class DiscBox(Gtk.ListBoxRow):
    """A widget which compromises one disc

    DiscBox contains a disc label for the disc number on top
    with a GtkListBox beneath.
    """
    __gtype_name__ = 'DiscBox'

    _list_box = Gtk.Template.Child()

    __gsignals__ = {
        'song-activated': (GObject.SignalFlags.RUN_FIRST, None, (Gtk.Widget,))
    }

    disc_nr = GObject.Property(type=int, default=1)

    def __init__(
            self, application: Application, corealbum: CoreAlbum,
            coredisc: CoreDisc) -> None:
        """Initialize

        :param Application coredisc: The Application object
        :param CoreAlbum corealbum: The corealbum of the coredisc
        :param CoreDisc coredisc: The CoreDisc object to use
        """
        super().__init__()

        self._application = application
        self._corealbum = corealbum
        self._coredisc = coredisc
        self._model: Gio.ListModel = coredisc.props.model

        self._coredisc.bind_property(
            "disc-nr", self, "disc-nr",
            GObject.BindingFlags.SYNC_CREATE)

        self._list_box.bind_model(self._model, self._create_widget)

    def _create_widget(self, coresong):
        song_widget = SongWidget(coresong)
        song_widget.props.menu = SongWidgetMenu(
            self._application, song_widget, self._corealbum)

        return song_widget

    @Gtk.Template.Callback()
    def _song_activated(
            self, list_box: Gtk.ListBox, song_widget: SongWidget) -> bool:
        self.emit("song-activated", song_widget)

        return Gdk.EVENT_STOP



# Copyright 2018 The GNOME Music developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

from enum import IntEnum

from gi.repository import Adw, GObject, Gtk


@Gtk.Template(resource_path="/org/gnome/Music/ui/HeaderBar.ui")
class HeaderBar(Adw.Bin):
    """Headerbar of the application"""

    class State(IntEnum):
        """States the Headerbar can have"""
        MAIN = 0
        CHILD = 1
        SEARCH = 2
        EMPTY = 3

    __gtype_name__ = "HeaderBar"

    _search_button = Gtk.Template.Child()
    _headerbar = Gtk.Template.Child()
    _menu_button = Gtk.Template.Child()

    search_mode_active = GObject.Property(type=bool, default=False)
    stack = GObject.Property(type=Adw.ViewStack)

    def __init__(self, application):
        """Initialize Headerbar

        :param Application application: Application object
        """
        super().__init__()

        self._stack_switcher = Adw.ViewSwitcher(
            focusable=False, halign="center",
            policy=Adw.ViewSwitcherPolicy.WIDE)

        self.bind_property(
            "stack", self._stack_switcher, "stack",
            GObject.BindingFlags.BIDIRECTIONAL
            | GObject.BindingFlags.SYNC_CREATE)
        self.bind_property(
            "search-mode-active", self._search_button, "active",
            GObject.BindingFlags.BIDIRECTIONAL
            | GObject.BindingFlags.SYNC_CREATE)

    @GObject.Property
    def state(self):
        """State of the widget

        :returns: Widget state
        :rtype: HeaderBar.State
        """
        return self._state

    @state.setter  # type: ignore
    def state(self, value):
        """Set state of the of widget

        This influences the look and functionality of the headerbar.

        :param HeaderBar.State value: Widget state
        """
        self._state = value
        self._update()

        search_visible = self.props.state != HeaderBar.State.SEARCH
        self._search_button.props.visible = search_visible

        if value == HeaderBar.State.EMPTY:
            self._search_button.props.sensitive = False
            self._stack_switcher.hide()
        else:
            self._search_button.props.sensitive = True
            self._stack_switcher.show()

    def _update(self):
        if self.props.state != HeaderBar.State.MAIN:
            self._headerbar.props.title_widget = None
        else:
            self._headerbar.props.title_widget = self._stack_switcher

        self._menu_button.props.visible = (
            self.props.state in [HeaderBar.State.MAIN, HeaderBar.State.EMPTY]
        )



# Copyright 2019 The GNOME Music developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

from __future__ import annotations
from typing import Any

from gettext import gettext as _
from gi.repository import Adw, GObject, Gtk

from gnomemusic.search import Search


@Gtk.Template(resource_path="/org/gnome/Music/ui/SearchHeaderBar.ui")
class SearchHeaderBar(Adw.Bin):
    """SearcnHeaderbar of the application"""

    __gtype_name__ = "SearchHeaderBar"

    _search_header_bar = Gtk.Template.Child()
    _search_button = Gtk.Template.Child()

    search_mode_active = GObject.Property(type=bool, default=False)
    search_state = GObject.Property(type=int, default=0)
    search_text = GObject.Property(type=str)

    def __init__(self, application):
        super().__init__()

        self._coregrilo = application.props.coregrilo

        self._entry = Gtk.SearchEntry()
        self._entry.props.placeholder_text = _(
            "Search songs, artists and albums")
        self._entry.props.halign = Gtk.Align.CENTER
        self._entry.props.visible = True
        self._entry.props.width_request = 500
        self._entry.props.search_delay = 250
        self._search_header_bar.props.title_widget = self._entry

        self.bind_property(
            "search-mode-active", self._search_button, "active",
            GObject.BindingFlags.BIDIRECTIONAL
            | GObject.BindingFlags.SYNC_CREATE)
        self.bind_property("search-text", self._entry, "text")
        self.connect("notify::search-text", self._on_search_text_changed)

        self.connect(
            "notify::search-mode-active", self._on_search_mode_changed)
        self.connect("notify::search-state", self._search_state_changed)

        self._entry.connect("search-changed", self._search_entry_changed)

    def _search_entry_changed(self, widget: Gtk.SearchEntry) -> bool:
        search_term = self._entry.get_text()
        self._coregrilo.search(search_term)

        if search_term == "":
            self._set_error_style(False)
            self.props.search_state = Search.State.NONE

        return False

    def _on_search_mode_changed(self, klass, data):
        if self.props.search_mode_active:
            self._entry.grab_focus()

    def _search_state_changed(self, klass, data):
        search_state = self.props.search_state

        if search_state == Search.State.NO_RESULT:
            self._set_error_style(True)
        elif search_state == Search.State.RESULT:
            self._set_error_style(False)
        elif search_state == Search.State.NONE:
            self._set_error_style(False)
            self._entry.props.text = ""

    def _set_error_style(self, error):
        """Adds error state to the search entry.

        :param bool error: Whether to add error state
        """
        style_context = self._entry.get_style_context()
        if error:
            style_context.add_class("error")
        else:
            style_context.remove_class("error")

    def _on_search_text_changed(
            self, searchheaderbar: SearchHeaderBar, data: Any) -> None:
        self._entry.set_position(len(self._entry.props.text))



# Copyright 2018 The GNOME Music developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

from __future__ import annotations
import typing

from gi.repository import GLib, GObject

if typing.TYPE_CHECKING:
    from gnomemusic.window import Window


class WindowPlacement(GObject.GObject):
    """Main window placement

    Restores and saves the main window placement, success may vary
    depending on the underlying window manager.
    """

    __gtype_name__ = 'WindowPlacement'

    def __init__(self, window):
        """Initialize WindowPlacement

        :param Gtk.Window window: Main window
        """
        super().__init__()

        self._window = window
        application = window.props.application
        self._settings = application.props.settings

        self._restore_window_state()

        self._window_placement_update_timeout = 0
        self._window.connect("notify::maximized", self._on_maximized)
        self._window.connect("notify::default-height", self._on_size_change)
        self._window.connect("notify::default-width", self._on_size_change)

    def _restore_window_state(self):
        size_setting = self._settings.get_value('window-size')
        if (len(size_setting) == 2
                and isinstance(size_setting[0], int)
                and isinstance(size_setting[1], int)):
            self._window.set_default_size(size_setting[0], size_setting[1])

        if self._settings.get_value('window-maximized'):
            self._window.maximize()

    def _on_size_change(
            self, window: Window, size: GObject.ParamSpecInt) -> None:
        if self._window_placement_update_timeout == 0:
            self._window_placement_update_timeout = GLib.timeout_add(
                500, self._store_size, window)

    def _store_size(self, window: Window) -> bool:
        width = window.get_width()
        height = window.get_height()
        self._settings.set_value(
            "window-size", GLib.Variant("ai", [width, height]))

        GLib.source_remove(self._window_placement_update_timeout)
        self._window_placement_update_timeout = 0

        return False

    def _on_maximized(self, klass, value, data=None):
        self._settings.set_boolean(
            'window-maximized', self._window.is_maximized())



# Copyright 2024 The GNOME Music developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from __future__ import annotations
from typing import Any
import typing

from gi.repository import Adw, Gtk

from gnomemusic.widgets.artistnavigationpage import ArtistNavigationPage
from gnomemusic.widgets.artistsearchtile import ArtistSearchTile
if typing.TYPE_CHECKING:
    from gnomemusic.coreartist import CoreArtist
    from gnomemusic.application import Application


@Gtk.Template(
    resource_path="/org/gnome/Music/ui/ArtistsSearchNavigationPage.ui")
class ArtistsSearchNavigationPage(Adw.NavigationPage):
    """ArtistsSearchNavigationPage
    """

    __gtype_name__ = "ArtistsSearchNavigationPage"

    _all_artists_flowbox = Gtk.Template.Child()

    def __init__(
            self, application: Application,
            model: Gtk.FlattenListModel) -> None:
        """Initialize the ArtistsSearchNavigationPage.

        :param Application application: The application object
        :param Gtk.FlattenListModel model: The model to show
        """
        super().__init__()

        self._application = application
        window = self._application.props.window
        self._navigation_view = window.props.navigation_view

        self._all_artists_flowbox.bind_model(
            model, self._create_artist_widget)

    def _create_artist_widget(
            self, coreartist: CoreArtist) -> ArtistSearchTile:
        artist_tile = ArtistSearchTile(coreartist)

        return artist_tile

    @Gtk.Template.Callback()
    def _on_artist_activated(
            self, flowbox: Gtk.FlowBox, tile: ArtistSearchTile,
            user_data: Any = None) -> None:
        coreartist = tile.props.coreartist

        artist_page = ArtistNavigationPage(self._application, coreartist)
        self._navigation_view.push(artist_page)
